import React from 'react';
import { CompositeLayer, HeatmapLayer } from 'deck.gl';
import HeatLayerOptions from './HeatLayerOptions';
import * as d3 from 'd3';
import { getDistance, getProperty } from '../../../utlis';
import { Edge, LayerProps } from '../../../mapvis.types';

/*
Potential use cases:
  - Logistics company
    Analyze the distribution of departure and arrival locations for shipments
    Areas with high concentrations of departures or arrivals can be visualized
    Optimize routes, allocate resources efficiently, and identify areas with higher demand for their services
  - Marketing company
    Analyzing graphs or networks that represent various attributes among entities
    Visualize social network interactions, with the color intensity showing the strength or frequency of interactions between users
    Identify key influencers, communities, or trends, enabling the marketing company to tailor their strategies based on these insights
  - Mobility company
    E.g. ride-sharing service or a public transportation provider
    Understanding the lengths of outgoing edges (distances) from specific nodes
    Highlight areas with longer average distances or higher travel times
    Guide decisions like where to position hubs, optimize vehicle placement
*/

export const typesMap = ['location', 'distance', 'attribute', 'path'];

export const locationMap = ['source', 'target'];

export const HeatLayerConfig = {
  type: 'location',
  location: 'source',
  entity: 'nodeAttributes',
  nSegments: 10,
  attribute: '',
};

export class HeatLayer extends CompositeLayer<LayerProps> {
  static type = 'Heatmap';
  static layerOptions = HeatLayerConfig;

  static generateLayerOptions(layer: any, updatedLayer: void, graphInfo: { [key: string]: any }, deleteLayer: void) {
    return <HeatLayerOptions layer={layer} updatedLayer={updatedLayer} graphInfo={graphInfo} deleteLayer={deleteLayer} />;
  }

  updateState({ props, oldProps, context, changeFlags }: { props: any; oldProps: any; context: any; changeFlags: any }) {
    console.log(props, oldProps, context, changeFlags);
  }

  createSegments(edges: Edge[]) {
    // Generates a path between source and target nodes
    return edges.map((edge: Edge, index) => {
      const length = getDistance(edge.path[0], edge.path[1]);
      const nSegments = length * this.props.config.nSegments;

      let xscale = d3
        .scaleLinear()
        .domain([0, nSegments + 1])
        .range([edge.path[0][0], edge.path[1][0]]);

      let yscale = d3
        .scaleLinear()
        .domain([0, nSegments + 1])
        .range([edge.path[0][1], edge.path[1][1]]);

      let source = edge.path[0];
      let target = null;
      let local = [source];

      for (let j = 1; j <= nSegments; j++) {
        target = [xscale(j), yscale(j)];
        local.push(target);
        source = target;
      }

      local.push(edge.path.slice(-1)[0]);
      return { ...edge, path: local };
    });
  }

  renderLayers() {
    const { graph, config, visible } = this.props;

    // if (!visible) return;

    const layers = [];

    if (config.type === 'location') {
      layers.push(
        new HeatmapLayer(
          this.getSubLayerProps({
            data:
              config.location === 'source'
                ? graph.getEdges().map((edge: Edge) => graph.getNode(edge.from))
                : graph.getEdges().map((edge: Edge) => graph.getNode(edge.to)),
            getPosition: (d: any) => [d.attributes.long, d.attributes.lat],
            aggregation: 'SUM',
          }),
        ),
      );
    } else if (config.type === 'distance') {
      layers.push(
        new HeatmapLayer(
          this.getSubLayerProps({
            data: graph.getEdges().map((edge: Edge) => {
              const from = graph.getNode(edge.from);
              const from_coords: [number, number] = [from.attributes.long, from.attributes.lat];
              const to = graph.getNode(edge.to);
              const to_coords: [number, number] = [to.attributes.long, to.attributes.lat];
              const dist = getDistance(from_coords, to_coords);
              const node = config.location === 'source' ? from : to;
              return { ...node, distance: dist };
            }),
            threshold: 0.5,
            getPosition: (d: any) => [d.attributes.long, d.attributes.lat],
            getWeight: (d: any) => d.distance,
            aggregation: 'MEAN',
          }),
        ),
      );
    } else if (config.type === 'attribute') {
      console.log('attribute');
      layers.push(
        new HeatmapLayer(
          this.getSubLayerProps({
            data: graph.getEdges().map((edge: Edge) => graph.getNode(edge.from)),
            getPosition: (d: any) => [d.attributes.long, d.attributes.lat],
            getWeight: (d: any) => {
              console.log(d, d.attributes[config.attribute]);
              return 1;
            },
            aggregation: 'SUM',
          }),
        ),
      );
    } else if (config.type === 'path') {
      layers.push(
        new HeatmapLayer(
          this.getSubLayerProps({
            data: this.createSegments(
              graph.getEdges().map((edge: Edge) => {
                return {
                  ...edge,
                  path: [this.props.graph.getNodeLocation(edge.from), this.props.graph.getNodeLocation(edge.to)],
                };
              }),
            ).flatMap((edge) => edge.path),
            getPosition: (d: any) => d,
            aggregation: 'SUM',
          }),
        ),
      );
    }

    return [...layers];
  }
}

HeatLayer.layerName = 'Heatmap';
