import { useEffect, useRef, useState, useMemo } from 'react';
import styles from './paohvis.module.scss';
import {
  Attribute,
  Relation,
  AttributeOrigin,
  EntitiesFromSchema,
  FilterInfo,
  NodeOrder,
  PaohvisData,
  PaohvisNodeOrder,
  RelationsFromSchema,
  ValueType,
} from './types';
import { useImmer } from 'use-immer';

import { getWidthOfText } from '../../../schema/schema-utils';
import { processDataColumn } from './utils/processAttributes';
import { select, selectAll, scaleOrdinal } from 'd3';
import { HyperEdgeRange } from './components/HyperEdgesRange';
import { ToPaohvisDataParserUseCase } from './utils/ToPaohvisDataParserUsecase';
import { MakePaohvisMenu } from './components/MakePaohvisMenu';
import { RowLabelColumn } from './components/RowLabelColumn';
import { GraphMetadata } from '@graphpolaris/shared/lib/data-access/statistics';
import { SettingsContainer } from '@graphpolaris/shared/lib/vis/components/config';
import { Input } from '@graphpolaris/shared/lib/components/inputs';
import { VisualizationPropTypes, VISComponentType } from '../../common';
import { isNodeLinkResult } from './utils/ResultNodeLinkParserUseCase';
import { calculateAttributesAndRelations, calculateAttributesFromRelation, calcTextWidthAndStringText } from './utils/utils';
import { visualizationColors } from 'config';

type PaohvisViewModelState = {
  //rowHeight: number;
  //hyperedgeColumnWidth: number;
  //gapBetweenRanges: number;

  hyperedgesOnRow: number[][];
  allHyperEdges: number[][];

  entitiesFromSchema: EntitiesFromSchema;
  relationsFromSchema: RelationsFromSchema;
  entityVertical: string;
  entityVerticalListed: string;
  entityHorizontal: string;
  chosenRelation: string;
  isEntityVerticalEqualToRelationFrom: boolean;
  nodeOrder: { orderBy: NodeOrder; isReverseOrder: boolean };
  paohvisFilters: { nodeFilters: FilterInfo[]; edgeFilters: FilterInfo[] };
  axisInfo: {
    selectedAttribute: {
      name: string;
      type: ValueType;
      origin: AttributeOrigin;
    };
    relation: { collection: string; from: string; to: string };
    isYAxisEntityEqualToRelationFrom: boolean;
  };
};

export type PaohVisProps = {
  rowHeight: number;
  hyperedgeColumnWidth: number;
  gapBetweenRanges: number;
  data?: PaohvisData;
  showColor: boolean;
};

const configuration: PaohVisProps = {
  rowHeight: 30,
  hyperedgeColumnWidth: 20,
  gapBetweenRanges: 5,
  showColor: false,
};

export const PaohVis = ({ data, schema, configuration }: VisualizationPropTypes) => {
  const graphQueryResult = data;

  const [isButtonPressed, setIsButtonPressed] = useState(false);

  const svgRef = useRef<SVGSVGElement>(null);
  const secondContainerRef = useRef<HTMLDivElement | null>(null);
  const dimensions = useRef({ width: 0, height: 0 });
  const [dataModel, setDataModel] = useState<PaohvisData>({
    rowLabelsList: [],
    rowLabels: [],
    hyperEdgeRanges: [],
    maxRowLabelWidth: 0,
  });

  const maxSizeTextColumns = 150;

  const colorRowsRef = useRef<{ [key: string]: string[] }>({});

  const getSecondContainerSize = () => {
    if (secondContainerRef.current) {
      const { offsetWidth, offsetHeight } = secondContainerRef.current;
      dimensions.current = { width: offsetWidth, height: offsetHeight };
    }
  };
  useEffect(() => {
    if (!svgRef.current) return;

    getSecondContainerSize();
  });

  // For each attribute that has less than a max of unique variables (number of colors available)
  // compute a color label for each row. So when the user needs it, that color can be applied.
  //const processedData = useMemo(() => {
  useMemo(() => {
    const attributesArray = graphQueryResult.nodes.map((node) => {
      const types =
        schema.nodes.find((n) => n.key === node.label)?.attributes?.attributes ??
        schema.edges.find((r) => r.key === node.label)?.attributes?.attributes ??
        [];

      return {
        attribute: node.attributes,
        type: Object.fromEntries(types.map((t: { name: string; type: string }) => [t.name, t.type])),
      };
    });

    const data2Render = Object.keys(attributesArray[0].attribute).map((dataColumn: string, i) => {
      return processDataColumn(39, dataColumn, attributesArray[0], attributesArray);
    });

    const numColorsAvailable = visualizationColors.GPCat.colors[14].length;
    const colorsAvailable = visualizationColors.GPCat.colors[14];

    const filteredArray = data2Render.filter((obj) => obj.numUniqueElements < numColorsAvailable);
    const colorMappings: { [name: string]: string[] } = {};

    filteredArray.forEach(({ name }) => {
      // gets categories for that object instance
      const dataExtracted = (data2Render.find((item) => item.name === name)?.data?.map((entry) => entry.category) || []) as any[];

      // assign that category to the node instance
      const nodeLabeled = dataModel.rowLabels.map(
        (id) => (graphQueryResult.nodes.find((item) => item.id === id) || {}).attributes?.[name] as string,
      );

      //const colorBase = scaleOrdinal<string>().domain(dataExtracted).range(['green', 'blue']); // use the defined colors
      const colorBase = scaleOrdinal<string>().domain(dataExtracted).range(colorsAvailable); // to test things
      const colorArray = nodeLabeled.map((value) => colorBase(value));
      colorMappings[name] = colorArray;
    });

    // add color base
    const resultArray = dataModel.rowLabels.map((_, index) => (index % 2 === 0 ? 'hsl(var(--clr-sec--300))' : 'hsl(var(--clr-sec--200))'));
    colorMappings['originalColor'] = resultArray;
    colorRowsRef.current = colorMappings;
  }, []);

  useEffect(() => {
    //console.log(colorRowsRef.current);
    if (configuration.showColor) {
      selectAll('.rowsLabel')
        .select('rect')
        .data(colorRowsRef.current.vip)
        //.attr('stroke', (d: string) => d)
        .attr('fill', (d: string) => d);

      selectAll('.hyperEdgeBlock')
        .selectAll('rect')
        .data(colorRowsRef.current.vip)
        //.attr('stroke', (d: string) => d)
        .attr('fill', (d: string) => d);
    } else {
      selectAll('.rowsLabel')
        .select('rect')
        .data(colorRowsRef.current.originalColor)
        //.attr('stroke', (d: string) => d)
        .attr('fill', (d: string) => d);

      selectAll('.hyperEdgeBlock')
        .selectAll('rect')
        .data(colorRowsRef.current.originalColor)
        //.attr('stroke', (d: string) => d)
        .attr('fill', (d: string) => d);
    }

    //
  }, [configuration.showColor]);

  // old props
  /*
    rowHeight: 30,
    hyperedgeColumnWidth: 30,
    gapBetweenRanges: 3,
  */

  // Change old props by configuration parameters. As percentages
  const props_rowHeight = Math.round(0.055 * dimensions.current.height * 10) / 10; //Math.round(0.03375 * dimensions.current.height * 10) / 10;
  const props_hyperedgeColumnWidth = Math.round(0.04 * dimensions.current.width * 10) / 10; //Math.round(0.01793 * dimensions.current.width * 10) / 10;
  const props_gapBetweenRanges = 0.0; //Math.round(0.001793 * dimensions.current.width * 10) / 10;

  const [viewModel, setViewModel] = useImmer<PaohvisViewModelState>({
    //rowHeight: 30,
    //hyperedgeColumnWidth: 20,
    //gapBetweenRanges: 5,

    hyperedgesOnRow: [],
    allHyperEdges: [],
    entitiesFromSchema: {
      entityNames: [],
      attributesPerEntity: {},
      relationsPerEntity: {},
    },
    relationsFromSchema: {
      relationCollection: [],
      relationNames: {},
      attributesPerRelation: {},
    },

    entityVertical: '',
    entityHorizontal: '',
    entityVerticalListed: '',
    chosenRelation: '',
    isEntityVerticalEqualToRelationFrom: true,
    nodeOrder: { orderBy: NodeOrder.degree, isReverseOrder: false },
    paohvisFilters: { nodeFilters: [], edgeFilters: [] },
    axisInfo: {
      selectedAttribute: {
        name: '',
        type: ValueType.noAttribute,
        origin: AttributeOrigin.noAttribute,
      },
      relation: { collection: '', from: '', to: '' },
      isYAxisEntityEqualToRelationFrom: true,
    },
  });
  // const [state, setState] = useState({
  //   entityVertical: '',
  //   entityHorizontal: '',
  //   relation: '',
  //   isEntityFromRelationFrom: true,
  // });

  //
  // Methods
  //

  /**
   * Filters the current PAOHvis visualization.
   * @param {boolean} isTargetEntity Tells if the the filter you want to remove is applied to an entity or relation.
   * @param {string} entityOrRelationType This is the type of the target.
   * @param {string} attribute This is the chosen attribute of the target you want to filter on.
   * @param {string} predicate This is the chosen predicate of how you want to compare the attribute.
   * @param {string} compareValue This is the chosen value which you want to compare the attribute values to.
   */
  function onClickFilterButton(
    isTargetEntity: boolean,
    entityOrRelationType: string,
    attribute: string,
    predicate: string,
    compareValue: string,
  ): void {
    const attributeName: string = attribute.split(':')[0];
    const attributeType: string = attribute.split(':')[1];
    const lowerCaseCompareValue = compareValue.toLowerCase();
    let compareValueTyped: any;

    // the compareValue must be typed based on the type of the attribute.
    switch (attributeType) {
      case ValueType.number:
        if (!Number.isNaN(Number(compareValue))) compareValueTyped = Number(compareValue);
        else throw new Error('Error: This is not a correct input');
        break;
      case ValueType.bool:
        if (lowerCaseCompareValue == 'true') compareValueTyped = true;
        else if (lowerCaseCompareValue == 'false') compareValueTyped = false;
        else throw new Error('Error: This is not a correct input');
        break;
      default:
        compareValueTyped = compareValue;
        break;
    }
    const newFilter: FilterInfo = {
      targetGroup: entityOrRelationType,
      attributeName: attributeName,
      value: compareValueTyped,
      predicateName: predicate,
    };

    if (isTargetEntity) {
      // The current filter is removed and then the new one is added.
      // TODO: to keep all filters, delete this line and improve config menu to see all filters instead of only the last one.
      setViewModel((draft) => {
        draft.paohvisFilters.nodeFilters.pop();
        draft.paohvisFilters.nodeFilters.push(newFilter);
        return draft;
      });
    } else {
      // The current filter is removed and then the new one is added.
      // TODO: to keep all filters, delete this line and improve config menu to see all filters instead of only the last one.
      setViewModel((draft) => {
        draft.paohvisFilters.edgeFilters.pop();
        draft.paohvisFilters.edgeFilters.push(newFilter);
        return draft;
      });
    }
  }

  /**
   * Resets the current chosen filter.
   * @param {boolean} isTargetEntity Tells if the the filter you want to remove is applied to an entity or relation.
   */
  function onClickResetButton(isTargetEntity: boolean): void {
    setViewModel((draft) => {
      if (isTargetEntity) draft.paohvisFilters.nodeFilters.pop();
      else draft.paohvisFilters.edgeFilters.pop();
      return draft;
    });
  }

  /**
   * Handles the visual changes when you enter a certain hyperEdge on hovering.
   * @param colIndex This is the index which states in which column you are hovering.
   * @param nameToShow This is the name of the entity which must be shown when you are hovering on this certain hyperEdge.
   */

  function onMouseEnterHyperEdge(colIndex: number, nameToShow: string, colTable: number): void {
    setViewModel((prevViewModel) => {
      // Calculate the changes using the latest state
      highlightAndFadeHyperEdges([colIndex]);
      highlightAndFadeRows(new Set(prevViewModel.allHyperEdges[colIndex]));
      showColName(colTable);

      // Return the new state
      return prevViewModel;
    });
  }

  /**
   * Handles the visual changes when you leave a certain hyperEdge on hovering.
   * @param colIndex This is the index which states in which column you were hovering.
   */
  function onMouseLeaveHyperEdge(colIndex: number): void {
    setViewModel((prevViewModel) => {
      unHighlightAndUnFadeHyperEdges([colIndex]);
      unHighlightAndUnFadeRows(new Set(prevViewModel.allHyperEdges[colIndex]));
      hideColName(colIndex);
    });
  }

  /**
   * This makes sure that the correct rows are highlighted and the correct rows are faded.
   * @param {Set<number>} rows This is the set with the numbers of the rows which must be highlighted.
   */
  function highlightAndFadeRows(rows: Set<number>) {
    rows.forEach((row) => highlightRow(row));

    const rowsToFade = [];
    for (let i = 0; i < dataModel.rowLabels.length; i++) if (!rows.has(i)) rowsToFade.push(i);
    rowsToFade.forEach((row) => fadeRow(row));
  }

  /**
   * This makes sure that the correct rows are unhighlighted and the correct rows are unfaded.
   * @param {Set<number>} rows This is the set with the numbers of the rows which must be unhighlighted.
   */
  function unHighlightAndUnFadeRows(rows: Set<number>) {
    rows.forEach((row) => unHighlightRow(row));

    const rowsToUnFade = [];
    for (let i = 0; i < dataModel.rowLabels.length; i++) if (!rows.has(i)) rowsToUnFade.push(i);
    rowsToUnFade.forEach((row) => unFadeRow(row));
  }

  /**
   * This makes sure that the correct hyperEdges are highlighted and the correct rows are faded.
   * @param {number[]} hyperEdges This is the list with the numbers of the hyperEdges which must be highlighted.
   */
  function highlightAndFadeHyperEdges(hyperEdges: number[]) {
    hyperEdges.forEach((hyperEdge) => highlightHyperedge(hyperEdge));

    const colsToFade = [];
    for (let i = 0; i < viewModel.allHyperEdges.length; i++) if (!hyperEdges.includes(i)) colsToFade.push(i);
    colsToFade.forEach((col) => fadeHyperedge(col));
  }

  /**
   * This makes sure that the correct hyperEdges are unhighlighted and the correct rows are unfaded.
   * @param {number[]} hyperEdges This is the list with the numbers of the hyperEdges which must be unhighlighted.
   */
  function unHighlightAndUnFadeHyperEdges(hyperEdges: number[]) {
    hyperEdges.forEach((hyperEdge) => unHighlightHyperedge(hyperEdge));

    const colsToUnFade = [];
    for (let i = 0; i < viewModel.allHyperEdges.length; i++) if (!hyperEdges.includes(i)) colsToUnFade.push(i);
    colsToUnFade.forEach((col) => unFadeHyperedge(col));
  }

  /**
   * This highlights one row given its index.
   * @param {number} rowIndex This is the index of the row that must be highlighted.
   */
  function highlightRow(rowIndex: number) {
    const row = selectAll('.row-' + rowIndex);
  }

  /**
   * This unhighlights one row given its index.
   * @param {number} rowIndex This is the index of the row that must be unhighlighted.
   */
  function unHighlightRow(rowIndex: number) {
    const row = selectAll('.row-' + rowIndex);
  }

  /**
   * This highlights one hyperEdge given its index.
   * @param {number} columnIndex This is the index of the hyperEdge that must be highlighted.
   */
  function highlightHyperedge(columnIndex: number) {
    const hyperedge = select('.col-' + columnIndex);
    hyperedge.selectAll('circle').style('fill', 'hsl(var(--a))');
  }

  /**
   * This unhighlights one hyperEdge given its index.
   * @param {number} columnIndex This is the index of the hyperEdge that must be unhighlighted.
   */
  function unHighlightHyperedge(columnIndex: number) {
    const hyperedge = select('.col-' + columnIndex);
    hyperedge.selectAll('circle').style('fill', 'hsl(var(--clr-white))');
  }

  /**
   * This fades one row given its index.
   * @param {number} rowIndex This is the index of the row that must be faded.
   */
  function fadeRow(rowIndex: number) {
    const row = selectAll('.row-' + rowIndex);
    row.select('text').attr('fill', 'hsl(var(--clr-sec--400))');
  }

  /**
   * This unfades one row given its index.
   * @param {number} rowIndex This is the index of the row that must be unfaded.
   */
  function unFadeRow(rowIndex: number) {
    const row = selectAll('.row-' + rowIndex);
    row.select('text').attr('fill', 'hsl(var(--clr-sec--800))'); //.attr('opacity', '1');
  }

  /**
   * This fades one hyperEdge given its index.
   * @param {number} columnIndex This is the index of the hyperEdge that must be faded.
   */
  function fadeHyperedge(columnIndex: number) {
    const hyperedge = select('.col-' + columnIndex);
    hyperedge.selectAll('circle').attr('opacity', '.3');
    hyperedge.selectAll('line').attr('opacity', '.3');
  }

  /**
   * This unfades one hyperEdge given its index.
   * @param {number} columnIndex This is the index of the hyperEdge that must be unfaded.
   */
  function unFadeHyperedge(columnIndex: number) {
    const hyperedge = select('.col-' + columnIndex);
    hyperedge.selectAll('circle').attr('opacity', '1');
    hyperedge.selectAll('line').attr('opacity', '1');
  }

  /**
   * This shows the name when hovering on a hyperEdge.
   * @param {string} nameToShow This is the name that must be shown.
   */
  function showColName(colNumber: number): void {
    //selectAll('.text-columns').select('text').attr('opacity', '.3');
    selectAll('.text-columns').select('text').attr('fill', 'hsl(var(--clr-sec--300))');
    // hsl(var(--clr-sec--800))

    select('.text-col-' + colNumber)
      .select('text')
      //.attr('opacity', '1');
      .attr('fill', 'hsl(var(--clr-sec--800))');
  }

  /** This hides the name when leaving a hyperEdge. */
  function hideColName(colNumber: number): void {
    selectAll('.text-columns').select('text').attr('fill', 'hsl(var(--clr-sec--800))');
  }

  /**
   * Makes the new PAOHvis visualization.
   * @param {string} entityVertical This is the name of the vertical entity (so on the left).
   * @param {string} entityHorizontal This is the name of the horizontal entity (so at the top).
   * @param {string} relationName This is the (collection)-name of the relation.
   * @param {boolean} isEntityVerticalEqualToRelationFrom Tells if the vertical entity is the from or to of the relation.
   * @param {Attribute} chosenAttribute This is the attribute on which the PAOHvis must be grouped by.
   * @param {PaohvisNodeOrder} nodeOrder Defines the sorting order of the PAOHvis visualization.
   */
  function onClickMakeButton(
    entityVertical: string,
    entityHorizontal: string,
    entityVerticalListed: string,
    relationName: string,
    isEntityVerticalEqualToRelationFrom: boolean,
    chosenAttribute: Attribute,
    nodeOrder: PaohvisNodeOrder,
  ): void {
    setViewModel((draft) => {
      draft.entityVertical = entityVertical;
      draft.entityVerticalListed = entityVerticalListed;
      draft.entityHorizontal = entityHorizontal;
      draft.chosenRelation = relationName;
      draft.isEntityVerticalEqualToRelationFrom = true; //isEntityVerticalEqualToRelationFrom;
      draft.nodeOrder = nodeOrder;

      return draft;
    });
    setAxisInfo(relationName, isEntityVerticalEqualToRelationFrom, chosenAttribute);
    setIsButtonPressed(true);
  }

  /**
   * This method parses data, makes a new Paohvis Table and lets the view re-render.
   */
  function makePaohvisTable() {
    // set new data
    const newData = new ToPaohvisDataParserUseCase(graphQueryResult).parseQueryResult(
      viewModel.axisInfo,
      viewModel.nodeOrder,
      dimensions,
      viewModel.entityVerticalListed,
      viewModel.entityHorizontal,
      viewModel.entityVertical,
    );
    setDataModel(newData);
  }

  /**
   * This method makes and sets a new PaohvisAxisInfo.
   * @param relationName is the relation that will be used in the Paohvis table.
   * @param isEntityVerticalEqualToRelationFrom is true when the entity on the y-axis belongs to the 'from' part of the relation.
   * @param chosenAttribute is the chosen attribute that will be used to divide the HyperEdges into HyperEdgeRanges.
   */
  function setAxisInfo(relationName: string, isEntityVerticalEqualToRelationFrom: boolean, chosenAttribute: Attribute): void {
    const namesOfEntityTypes = viewModel.relationsFromSchema.relationNames[relationName].split(':');
    const relation: Relation = {
      collection: relationName,
      from: namesOfEntityTypes[0],
      to: namesOfEntityTypes[1],
    };

    // set axisInfo
    setViewModel((draft) => {
      draft.axisInfo = {
        selectedAttribute: chosenAttribute,
        relation: relation,
        isYAxisEntityEqualToRelationFrom: true, //isEntityVerticalEqualToRelationFrom,
      };
      return draft;
    });
  }

  /**
   * Handles the visual changes when you enter a certain row on hovering.
   * @param {number} rowIndex This is the index which states in which row you are hovering.
   */

  function onMouseEnterRow(rowIndex: number): void {
    setViewModel((prevViewModel) => {
      const rowsToHighlight = new Set<number>();
      const colsToHighlight: number[] = [];

      prevViewModel.hyperedgesOnRow[rowIndex].forEach((hyperedge) => {
        colsToHighlight.push(hyperedge);

        prevViewModel.allHyperEdges[hyperedge].forEach((row) => {
          rowsToHighlight.add(row);
        });
      });

      // Use the latest state to calculate the changes
      highlightAndFadeRows(rowsToHighlight);
      highlightAndFadeHyperEdges(colsToHighlight);

      // Return the new state
      return prevViewModel;
    });
  }

  /**
   * Handles the visual changes when you leave a certain row on hovering.
   * @param {number} rowIndex This is the index which states in which row you were hovering.
   */

  function onMouseLeaveRow(rowIndex: number): void {
    setViewModel((prevViewModel) => {
      const rowsToUnHighlight = new Set<number>();
      const colsToUnHighlight: number[] = [];

      prevViewModel.hyperedgesOnRow[rowIndex].forEach((hyperedge) => {
        colsToUnHighlight.push(hyperedge);

        prevViewModel.allHyperEdges[hyperedge].forEach((row) => {
          rowsToUnHighlight.add(row);
        });
      });

      // Use the latest state to calculate the changes
      unHighlightAndUnFadeRows(rowsToUnHighlight);
      unHighlightAndUnFadeHyperEdges(colsToUnHighlight);

      // Return the new state
      return prevViewModel;
    });
  }

  //
  // Reactivity
  //

  useEffect(() => {
    setViewModel((draft) => {
      // When a schema is received, extract the entity names, and attributes per data type
      draft.entitiesFromSchema = calculateAttributesAndRelations(schema);
      draft.relationsFromSchema = calculateAttributesFromRelation(schema);
      return draft;
    });
  }, [schema]);

  /** This method filters and makes a new Paohvis table. */
  useEffect(() => {
    if (isNodeLinkResult(graphQueryResult)) {
      makePaohvisTable();
    } else {
      console.error('Invalid query result!');
    }
  }, [viewModel.paohvisFilters, graphQueryResult, viewModel.axisInfo, viewModel.nodeOrder]);

  useEffect(() => {
    /**
     * Recalculates the hyperEdges and hyperEdgeRanges for the state.
     */
    setViewModel((draft) => {
      // Create connecteHyperedgesInColumns, to quickly lookup which hyperedges are on a row
      draft.allHyperEdges = [];

      dataModel.hyperEdgeRanges.forEach((hyperEdgeRange) => {
        hyperEdgeRange.hyperEdges.forEach((hyperEdge) => {
          draft.allHyperEdges.push(hyperEdge.indices);
        });
      });

      draft.hyperedgesOnRow = dataModel.rowLabels.map(() => []);
      draft.allHyperEdges.forEach((hyperedge, i) => {
        hyperedge.forEach((row) => {
          draft.hyperedgesOnRow[row].push(i);
        });
      });
      return draft;
    });
  }, [dataModel]);

  //
  // RENDER
  //
  //

  const hyperEdgeRanges = dataModel.hyperEdgeRanges;
  const rowLabelColumnWidth = dataModel.maxRowLabelWidth;
  const hyperedgeColumnWidth = configuration.hyperedgeColumnWidth;

  //calculate yOffset - hyperedges
  /*
  let maxColWidth = 0;
  hyperEdgeRanges.forEach((hyperEdgeRange) => {
    const textLength = getWidthOfText(hyperEdgeRange.rangeText, styles.tableFontFamily, styles.tableFontSize, styles.tableFontWeight);
    console.log(textLength, hyperEdgeRange.rangeText);
    if (textLength > maxColWidth) maxColWidth = textLength;
  });
  */
  const yOffset = maxSizeTextColumns;
  const expandButtonWidth = 0.0; //parseFloat(styles.expandButtonSize.split('px')[0]);
  const margin = 0.0; //Math.round(100 * 0.00298 * dimensions.current.width) / 100; //5;

  // text
  const classTopTextColums = 'font-inter font-secondary font-semibold font-base';
  const textTopColumns = hyperEdgeRanges.map((item) => item.rangeText);
  const textTopColumnModifiedAndWidth = calcTextWidthAndStringText(textTopColumns, 0.8 * maxSizeTextColumns, classTopTextColums);
  const textTopColumnModified = textTopColumnModifiedAndWidth.map((item) => item.truncatedText);

  //calc table width
  let tableWidth = 0;
  let tableWidthWithExtraColumnLabelWidth = 0;
  hyperEdgeRanges.forEach((hyperEdgeRange) => {
    // size text
    const columnLabelWidth =
      /*Math.cos(columnLabelAngleInRadians) * */
      getWidthOfText(hyperEdgeRange.rangeText, styles.tableFontFamily, styles.tableFontSize, styles.tableFontWeight);

    const columnWidth = hyperEdgeRange.hyperEdges.length * hyperedgeColumnWidth + configuration.gapBetweenRanges * 3;
    // new text compute

    tableWidth += columnWidth;

    if (columnLabelWidth > columnWidth) {
      const currentTableWidthWithLabel = tableWidth + columnLabelWidth;
      if (currentTableWidthWithLabel > tableWidthWithExtraColumnLabelWidth)
        tableWidthWithExtraColumnLabelWidth = currentTableWidthWithLabel;
    }
    if (tableWidth > tableWidthWithExtraColumnLabelWidth) tableWidthWithExtraColumnLabelWidth = tableWidth;
  });
  tableWidthWithExtraColumnLabelWidth += rowLabelColumnWidth + expandButtonWidth + margin;
  //make  all hyperEdgeRanges
  let colOffset = 0;
  let hyperEdgeRangeColumns: JSX.Element[] = [];
  hyperEdgeRanges.forEach((hyperEdgeRange, i) => {
    hyperEdgeRangeColumns.push(
      <HyperEdgeRange
        key={i}
        index={i}
        text={textTopColumnModified[i]}
        hyperEdges={hyperEdgeRange.hyperEdges}
        nRows={dataModel.rowLabels.length}
        nCols={dataModel.hyperEdgeRanges.length}
        colOffset={colOffset}
        xOffset={rowLabelColumnWidth}
        yOffset={yOffset}
        rowHeight={configuration.rowHeight}
        textBothMargins={0.1 * maxSizeTextColumns}
        hyperedgeColumnWidth={hyperedgeColumnWidth}
        gapBetweenRanges={configuration.gapBetweenRanges}
        onMouseEnter={onMouseEnterHyperEdge}
        onMouseLeave={onMouseLeaveHyperEdge}
      />,
    );
    colOffset += hyperEdgeRange.hyperEdges.length;
  });

  // returns the whole PAOHvis visualization panel
  return (
    <div className={styles.container}>
      <MakePaohvisMenu makePaohvis={onClickMakeButton} graphQueryResult={graphQueryResult} schema={schema} />
      <div className="w-full h-full">
        <div className="w-full h-full">
          <div className="w-full h-full flex flex-row justify-center" ref={secondContainerRef}>
            <svg
              ref={svgRef}
              style={{
                width: tableWidthWithExtraColumnLabelWidth,
                height: yOffset + (dataModel.rowLabels.length + 1) * configuration.rowHeight,
              }}
            >
              <RowLabelColumn
                onMouseEnter={onMouseEnterRow}
                onMouseLeave={onMouseLeaveRow}
                titles={dataModel.rowLabels}
                width={rowLabelColumnWidth}
                rowHeight={configuration.rowHeight} // viewModel.rowHeight?
                yOffset={yOffset}
              />
              {hyperEdgeRangeColumns}
            </svg>
          </div>
        </div>
      </div>
    </div>
  );
};

const PaohSettings = ({
  configuration,
  graph,
  updateSettings,
}: {
  configuration: PaohVisProps;
  graph: GraphMetadata;
  updateSettings: (val: any) => void;
}) => {
  return (
    <SettingsContainer>
      <Input type="number" label="Row height" value={configuration.rowHeight} onChange={(val) => updateSettings({ rowHeight: val })} />
      <Input
        type="number"
        label="Hyper edge column width"
        value={configuration.hyperedgeColumnWidth}
        onChange={(val) => updateSettings({ hyperedgeColumnWidth: val })}
      />
      <Input
        type="number"
        label="Gap between ranges"
        value={configuration.gapBetweenRanges}
        onChange={(val) => updateSettings({ gapBetweenRanges: val })}
      />
    </SettingsContainer>
  );
};

export const PaohVisComponent: VISComponentType = {
  displayName: 'PaohVis',
  component: PaohVis,
  settings: PaohSettings,
  configuration: configuration,
};

export default PaohVisComponent;
